package ru.dbotthepony.kstarbound.util

import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.io.readDouble
import ru.dbotthepony.kstarbound.io.writeDouble
import ru.dbotthepony.kstarbound.network.syncher.NetworkedElement
import java.io.DataInputStream
import java.io.DataOutputStream

interface IClock {
	val nanos: Long
	val micros: Long get() = nanos / 1_000L
	val millis: Long get() = nanos / 1_000_000L
	val seconds: Double get() = (nanos / 1_000L) / 1_000_000.0
}

// this is stupid, but legacy protocol requires it
// Used for timing in-game events which must be persistent
// And some worlds are not persistent (they get their own clocks,
// such as instance worlds, e.g. Outpost, or Creon Embassy from Elithian Races mod)

// https://www.pcgamingwiki.com/wiki/Category:Persistent describes it as:
// Gameplay continues even when player is not playing the game,
// and the game state is either simulated on a remote server or
// changes over time are calculated when the player returns to the game.
class RelativeClock() : IClock {
	constructor(stream: DataInputStream, isLegacy: Boolean) : this() {
		read(stream, isLegacy)
	}

	private var pointOfReference = 0L
	private var pointOfReferenceSet = false

	override var nanos: Long = 0L
		private set

	fun set(age: Long) {
		pointOfReferenceSet = false
		nanos = age
	}

	fun update(newPointOfReference: Long) {
		if (pointOfReferenceSet) {
			val diff = newPointOfReference - pointOfReference

			if (diff > 0L)
				nanos += diff
		}

		pointOfReference = newPointOfReference
	}

	fun update(newPointOfReference: Double) {
		return update((newPointOfReference * 1_000_000_000.0).toLong())
	}

	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeBoolean(pointOfReferenceSet)

		if (isLegacy) {
			if (pointOfReferenceSet)
				stream.writeDouble(pointOfReference / 1_000_000_000.0)

			stream.writeDouble(nanos / 1_000_000_000.0)
		} else {
			if (pointOfReferenceSet)
				stream.writeLong(pointOfReference)

			stream.writeLong(nanos)
		}
	}

	fun read(stream: DataInputStream, isLegacy: Boolean) {
		pointOfReferenceSet = stream.readBoolean()

		if (isLegacy) {
			if (pointOfReferenceSet)
				pointOfReference = (stream.readDouble() * 1_000_000_000.0).toLong()

			nanos = (stream.readDouble() * 1_000_000_000.0).toLong()
		} else {
			if (pointOfReferenceSet)
				pointOfReference = stream.readLong()

			nanos = stream.readLong()
		}
	}
}

class JVMClock : IClock {
	var origin = System.nanoTime()
		private set

	var baseline = 0L
		private set

	var isPaused = false
		private set

	fun set(nanos: Long) {
		origin = System.nanoTime()
		baseline = nanos
	}

	fun pause() {
		if (!isPaused) {
			baseline += System.nanoTime() - origin
			isPaused = true
		}
	}

	fun unpause() {
		if (isPaused) {
			origin = System.nanoTime()
			isPaused = false
		}
	}

	override val nanos: Long
		get() = if (isPaused) baseline else (System.nanoTime() - origin) + baseline
}

class GameTimer(val time: Double = 0.0) {
	constructor(stream: DataInputStream, isLegacy: Boolean) : this(stream.readDouble(isLegacy)) {
		timer = stream.readDouble(isLegacy)
	}

	fun write(stream: DataOutputStream, isLegacy: Boolean) {
		stream.writeDouble(time, isLegacy)
		stream.writeDouble(timer, isLegacy)
	}

	var timer = time
		private set

	fun reset() {
		timer = time
	}

	var hasFinished: Boolean
		get() = timer <= 0.0
		set(value) {
			if (value)
				timer = 0.0
			else
				timer = time
		}

	val percent: Double
		get() = if (time != 0.0) timer / time else 0.0

	fun invert() {
		timer = time - timer
	}

	fun tick(delta: Double = Starbound.TIMESTEP): Boolean {
		timer = (timer - delta).coerceAtLeast(0.0)
		return timer == 0.0
	}

	fun wrapTick(delta: Double = Starbound.TIMESTEP): Boolean {
		val result = tick(delta)
		if (result) reset()
		return result
	}
}
